import logging
from typing import List, Union

from pyspark.sql import DataFrame

_logger = logging.getLogger(__name__)


def standardize_checkpoint_location(checkpoint_location):
    if checkpoint_location is None:
        return checkpoint_location
    checkpoint_location = checkpoint_location.strip()
    if checkpoint_location == "":
        checkpoint_location = None
    return checkpoint_location


def _is_spark_connect_data_frame(df):
    # We cannot directly pyspark.sql.connect.dataframe.DataFrame as it requires Spark 3.4, which
    # is not installed on DBR 12.2 and earlier. Instead, we string match on the type.
    return (
        type(df).__name__ == "DataFrame"
        and type(df).__module__ == "pyspark.sql.connect.dataframe"
    )


def check_dataframe_type(df):
    """
    Check if df is a PySpark DataFrame, otherwise raise an error.
    """
    if not (isinstance(df, DataFrame) or _is_spark_connect_data_frame(df)):
        raise ValueError(
            f"Unsupported DataFrame type: {type(df)}. DataFrame must be a PySpark DataFrame."
        )


def check_kwargs_empty(the_kwargs, method_name):
    if len(the_kwargs) != 0:
        raise TypeError(
            f"{method_name}() got unexpected keyword argument(s): {list(the_kwargs.keys())}"
        )


def check_duplicate_keys(keys: Union[str, List[str]], key_name: str) -> None:
    """
    Check if there are duplicate keys. Raise an error if there is duplicates.
    """
    if keys and isinstance(keys, list):
        seen = set()
        for k in keys:
            if k in seen:
                raise ValueError(
                    f"Found duplicated key '{k}' in {key_name}. {key_name} must be unique."
                )
            seen.add(k)



"""
Wedata FeatureStoreClient Python实现
"""

from __future__ import annotations

from types import ModuleType
from typing import Union, List, Dict, Optional, Any
from pyspark.sql import DataFrame, SparkSession
from pyspark.sql.streaming import StreamingQuery
from pyspark.sql.types import StructType
import mlflow

from wedata.feature_store.constants.constants import APPEND, DEFAULT_WRITE_STREAM_TRIGGER
from wedata.feature_store.entities.feature_function import FeatureFunction
from wedata.feature_store.entities.feature_lookup import FeatureLookup
from wedata.feature_store.entities.feature_table import FeatureTable
from wedata.feature_store.entities.training_set import TrainingSet
from wedata.feature_store.feature_table_client.feature_table_client import FeatureTableClient
from wedata.feature_store.spark_client.spark_client import SparkClient
from wedata.feature_store.training_set_client.training_set_client import TrainingSetClient
from wedata.feature_store.utils import common_utils
from wedata.feature_store.utils.feature_utils import format_feature_lookups_and_functions


class FeatureStoreClient:
    """特征存储统一客户端，提供特征全生命周期管理能力"""

    def __init__(self, spark: SparkSession):
        """
        :param spark: 已初始化的SparkSession对象
        """

        self._spark = spark
        self._spark_client = SparkClient(spark)
        self._feature_table_client = FeatureTableClient(spark)
        self._training_set_client = TrainingSetClient(self._spark_client)

    def create_table(
            self,
            name: str,
            primary_keys: Union[str, List[str]],
            df: Optional[DataFrame] = None,
            *,
            timestamp_keys: Union[str, List[str], None] = None,
            partition_columns: Union[str, List[str], None] = None,
            schema: Optional[StructType] = None,
            description: Optional[str] = None,
            tags: Optional[Dict[str, str]] = None
    ) -> FeatureTable:
        """
        创建特征表（支持批流数据写入）

        Args:
            name: 特征表全称（格式：<table>）
            primary_keys: 主键列名（支持复合主键）
            df: 初始数据（可选，用于推断schema）
            timestamp_keys: 时间戳键（用于时态特征）
            partition_columns: 分区列（优化存储查询）
            schema: 表结构定义（可选，当不提供df时必需）
            description: 业务描述
            tags: 业务标签

        Returns:
            FeatureTable实例

        Raises:
            ValueError: 当schema与数据不匹配时
        """

        return self._feature_table_client.create_table(
            name=name,
            primary_keys=primary_keys,
            df=df,
            timestamp_keys=timestamp_keys,
            partition_columns=partition_columns,
            schema=schema,
            description=description,
            tags=tags
        )

    def read_table(self, name: str) -> DataFrame:
        """
        读取特征表数据

        Args:
            name: 特征表名称

        Returns:
            DataFrame: 包含特征表数据的DataFrame对象
        """

        return self._feature_table_client.read_table(name)

    def get_table(self, name: str) -> FeatureTable:
        """
        获取特征表元数据
        Args:
            name: 特征表名称

        Returns:
            FeatureTable: 包含特征表元数据的FeatureTable对象
        """

        return self._feature_table_client.get_table(name, self._spark_client)

    def drop_table(self, name: str) -> None:
        """
        删除特征表

        Args:
            name: 要删除的特征表名称

        Returns:
            None
        """

        return self._feature_table_client.drop_table(name)

    def write_table(
            self,
            name: str,
            df: DataFrame,
            mode: str = APPEND,
            checkpoint_location: Optional[str] = None,
            trigger: Dict[str, Any] = DEFAULT_WRITE_STREAM_TRIGGER,
    ) -> Optional[StreamingQuery]:
        """
        写入数据到特征表（支持批处理和流式处理）

        Args:
            name: 特征表名称
            df: 要写入的数据DataFrame
            mode: 写入模式（默认追加）
            checkpoint_location: 流式处理的检查点位置（可选）
            trigger: 流式处理触发器配置（默认使用系统预设）

        Returns:
            如果是流式写入返回StreamingQuery对象，否则返回None
        """

        return self._feature_table_client.write_table(
            name=name,
            df=df,
            mode=mode,
            checkpoint_location=checkpoint_location,
            trigger=trigger,
        )

    def create_training_set(
            self,
            df: DataFrame,
            feature_lookups: List[Union[FeatureLookup, FeatureFunction]],
            label: Union[str, List[str], None],
            exclude_columns: Optional[List[str]] = None,
            **kwargs,
    ) -> TrainingSet:
        """
        创建训练集

        Args:
            df: 基础数据
            feature_lookups: 特征查询列表
            label: 标签列名
            exclude_columns: 排除列名

        Returns:
            TrainingSet实例
        """

        if exclude_columns is None:
            exclude_columns = []

        # 如果为FeatureLookup，则将需要校验FeatureLookup的table_name，并构建完整表名
        for feature in feature_lookups:
            if isinstance(feature, FeatureLookup):
                if not feature.table_name:
                    raise ValueError("FeatureLookup must specify a table_name")
                # 先校验表名格式是否合法
                common_utils.validate_table_name(feature.table_name)
                # 再构建完整表名
                feature.table_name = common_utils.build_full_table_name(feature.table_name)

        features = feature_lookups
        del feature_lookups

        features = format_feature_lookups_and_functions(self._spark_client, features)

        return self._training_set_client.create_training_set_from_feature_lookups(
            df=df,
            feature_lookups=features,
            label=label,
            exclude_columns=exclude_columns,
            **kwargs
        )

    def log_model(
            self,
            model: Any,
            artifact_path: str,
            *,
            flavor: ModuleType,
            training_set: Optional[TrainingSet] = None,
            registered_model_name: Optional[str] = None,
            await_registration_for: int = mlflow.tracking._model_registry.DEFAULT_AWAIT_MAX_SLEEP_SECONDS,
            infer_input_example: bool = False,
            **kwargs,
    ):
        """
         记录MLflow模型并关联特征查找信息

         注意：必须使用TrainingSet.load_df返回的DataFrame训练模型，
         任何对DataFrame的修改(如标准化、添加列等)都不会在推理时应用

         Args:
             model: 要记录的模型对象
             artifact_path: 模型存储路径
             flavor: MLflow模型类型模块(如mlflow.sklearn)
             training_set: 训练模型使用的TrainingSet对象(可选)
             registered_model_name: 要注册的模型名称(可选)
             await_registration_for: 等待模型注册完成的秒数(默认300秒)
             infer_input_example: 是否自动记录输入示例(默认False)

         Returns:
             None
         """

        self._training_set_client.log_model(
            model=model,
            artifact_path=artifact_path,
            flavor=flavor,
            training_set=training_set,
            registered_model_name=registered_model_name,
            await_registration_for=await_registration_for,
            infer_input_example=infer_input_example,
            **kwargs
        )

    @property
    def spark(self):
        return self._spark



from collections import defaultdict
from typing import List

from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.catalog import Column
from pyspark.sql.functions import when, isnull
from pyspark.sql.types import StructType, StringType, StructField

from wedata.feature_store.entities.feature import Feature
from wedata.feature_store.entities.feature_table import FeatureTable
from wedata.feature_store.entities.function_info import FunctionParameterInfo, FunctionInfo
from wedata.feature_store.utils.common_utils import unsanitize_identifier, sanitize_multi_level_name


class SparkClient:
    def __init__(self, spark: SparkSession):
        self._spark = spark


    def get_current_catalog(self):
        """
        获取当前Spark会话的catalog名称（使用spark.catalog.currentCatalog属性）

        返回:
            str: 当前catalog名称，如果未设置则返回None
        """
        try:
            return unsanitize_identifier(self._spark.catalog.currentCatalog())
        except Exception:
            return None

    def get_current_database(self):
        """
        获取Spark上下文中当前设置的数据库名称

        返回:
            str: 当前数据库名称，如果获取失败则返回None
        """
        try:
            # 使用Spark SQL查询当前数据库
            df = self._spark.sql("SELECT CURRENT_DATABASE()")
            # 获取第一行第一列的值并去除特殊字符
            return unsanitize_identifier(df.first()[0])
        except Exception:
            # 捕获所有异常并返回None
            return None




    def createDataFrame(self, data, schema) -> DataFrame:
        return self._spark.createDataFrame(data, schema)


    def read_table(self, table_name):
        """读取Spark表数据

        Args:
            table_name: 表名，支持格式: catalog.schema.table、schema.table 或 table

        Returns:
            DataFrame: 表数据

        Raises:
            ValueError: 当表不存在或读取失败时抛出
        """
        try:
            # 解析表名
            parts = table_name.split('.')
            if len(parts) == 3:
                catalog, schema, table = parts
            elif len(parts) == 2:
                schema, table = parts
            else:
                table = table_name

            # 验证表是否存在
            if not self._spark.catalog.tableExists(table):
                raise ValueError(f"表不存在: {table_name}")

            return self._spark.table(table)

        except Exception as e:
            raise ValueError(f"读取表 {table_name} 失败: {str(e)}")


    def get_features(self, table_name):
        # 解析表名
        parts = table_name.split('.')
        if len(parts) == 3:
            # 对于三部分名称(catalog.schema.table)，使用schema.table格式
            _, schema, table = parts
            full_table_name = f"{schema}.{table}"
        elif len(parts) == 2:
            # 对于两部分名称(schema.table)，直接使用
            full_table_name = table_name
        else:
            # 单表名，使用当前数据库
            current_db = self.get_current_database()
            if not current_db:
                raise ValueError("无法确定当前数据库")
            full_table_name = f"{current_db}.{table_name}"

        # 使用dbName.tableName格式查询列信息
        columns = self._spark.catalog.listColumns(tableName=full_table_name)
        return [
            Feature(
                feature_table=table_name,
                feature_id=f"{table_name}_{row.name}",
                name=row.name,
                data_type=row.dataType,
                description=row.description or ""
            ) for row in columns
        ]

    def get_feature_table(self, table_name):

        # 获取表元数据
        table = self._spark.catalog.getTable(table_name)

        parts = table_name.split('.')
        if len(parts) == 3:
            # 对于三部分名称(catalog.schema.table)，只使用表名部分
            table_to_describe = parts[2]
        elif len(parts) == 2:
            # 对于两部分名称(schema.table)，只使用表名部分
            table_to_describe = parts[1]
        else:
            # 单表名，直接使用
            table_to_describe = table_name
        # 获取表详细信息
        table_details = self._spark.sql(f"DESCRIBE TABLE EXTENDED {table_to_describe}").collect()

        table_properties = {}
        for row in table_details:
            if row.col_name == "Table Properties":
                props = row.data_type[1:-1].split(", ")
                table_properties = dict(p.split("=") for p in props if "=" in p)

        # 获取特征列信息
        features = self.get_features(table_name)

        # 构建完整的FeatureTable对象
        return FeatureTable(
            name=table_name,
            table_id=table_properties.get("table_id", table_name),
            description=table.description or table_properties.get("comment", table_name),
            primary_keys=table_properties.get("primaryKeys", "").split(",") if table_properties.get("primaryKeys") else [],
            partition_columns=table.partitionColumnNames if hasattr(table, 'partitionColumnNames') else [],
            features=features,
            creation_timestamp=None,  # Spark表元数据不包含创建时间戳
            online_stores=None,
            notebook_producers=None,
            job_producers=None,
            table_data_sources=None,
            path_data_sources=None,
            custom_data_sources=None,
            timestamp_keys=table_properties.get("timestamp_keys"),
            tags=table_properties
        )

    def _get_routines_with_parameters(self, full_routine_names: List[str]) -> DataFrame:
        """
        Retrieve the routines with their parameters from information_schema.routines, information_schema.parameters.
        Return DataFrame only contains routines that 1. exist and 2. the caller has GetFunction permission on.

        Note: The returned DataFrame contains the cartesian product of routines and parameters.
        For efficiency, routines table columns are only present in the first row for each routine.
        """
        routine_name_schema = StructType(
            [
                StructField("specific_catalog", StringType(), False),
                StructField("specific_schema", StringType(), False),
                StructField("specific_name", StringType(), False),
            ]
        )
        routine_names_df = self.createDataFrame(
            [full_routine_name.split(".") for full_routine_name in full_routine_names],
            routine_name_schema,
        )
        routines_table = self.read_table(
            "system.information_schema.routines"
        )
        parameters_table = self.read_table(
            "system.information_schema.parameters"
        )

        # Inner join routines table to filter out non-existent routines.
        # Left join parameters as routines may have no parameters.
        full_routines_with_parameters_df = routine_names_df.join(
            routines_table, on=routine_names_df.columns, how="inner"
        ).join(parameters_table, on=routine_names_df.columns, how="left")

        # Return only relevant metadata from information_schema, sorted by routine name + parameter order.
        # For efficiency, only preserve routine column values in the first of each routine's result rows.
        # The first row will have parameter.ordinal_value is None (no parameters) or equals 0 (first parameter).
        def select_if_first_row(col: Column) -> Column:
            return when(
                isnull(parameters_table.ordinal_position)
                | (parameters_table.ordinal_position == 0),
                col,
                ).otherwise(None)

        return full_routines_with_parameters_df.select(
            routine_names_df.columns
            + [
                select_if_first_row(routines_table.routine_definition).alias(
                    "routine_definition"
                ),
                select_if_first_row(routines_table.external_language).alias(
                    "external_language"
                ),
                parameters_table.ordinal_position,
                parameters_table.parameter_name,
                parameters_table.full_data_type,
            ]
        ).sort(routine_names_df.columns + [parameters_table.ordinal_position])

    def get_functions(self, full_function_names: List[str]) -> List[FunctionInfo]:
        """
        Retrieves and maps Unity Catalog functions' metadata as FunctionInfos.
        """
        # Avoid unnecessary Spark calls and return if empty.
        if not full_function_names:
            return []

        # Collect dict of routine name -> DataFrame rows describing the routine.
        routines_with_parameters_df = self._get_routines_with_parameters(
            full_routine_names=full_function_names
        )
        routine_infos = defaultdict(list)
        for r in routines_with_parameters_df.collect():
            routine_name = f"{r.specific_catalog}.{r.specific_schema}.{r.specific_name}"
            routine_infos[routine_name].append(r)

        # Mock GetFunction DNE error, since information_schema does not throw.
        for function_name in full_function_names:
            if not function_name in routine_infos:
                raise ValueError(f"Function '{function_name}' does not exist.")

        # Map routine_infos into FunctionInfos.
        function_infos = []
        for function_name in full_function_names:
            routine_info = routine_infos[function_name][0]
            input_params = [
                FunctionParameterInfo(name=p.parameter_name, type_text=p.full_data_type)
                for p in routine_infos[function_name]
                if p.ordinal_position is not None
            ]
            function_infos.append(
                FunctionInfo(
                    full_name=function_name,
                    input_params=input_params,
                    routine_definition=routine_info.routine_definition,
                    external_language=routine_info.external_language,
                )
            )
        return function_infos





from collections import defaultdict
from typing import List, Optional

from pyspark.sql import Column, DataFrame
from pyspark.sql.functions import isnull, when
from pyspark.sql.types import StringType, StructField, StructType

class FunctionParameterInfo():
    def __init__(self, name: str, type_text: str):
        self._name = name
        self._type_text = type_text

    @property
    def name(self) -> str:
        return self._name

    @property
    def type_text(self) -> str:
        return self._type_text

    @classmethod
    def from_dict(cls, function_parameter_info_json):
        return FunctionParameterInfo(
            function_parameter_info_json["name"],
            function_parameter_info_json["type_text"],
        )


class FunctionInfo():
    """
    Helper entity class that exposes properties in GetFunction's response JSON as attributes.
    https://docs.databricks.com/api-explorer/workspace/functions/get

    Note: empty fields (e.g. when 0 input parameters) are not included in the response JSON.
    """

    # Python UDFs have external_language = "Python"
    PYTHON = "Python"

    def __init__(
        self,
        full_name: str,
        input_params: List[FunctionParameterInfo],
        routine_definition: Optional[str],
        external_language: Optional[str],
    ):
        self._full_name = full_name
        self._input_params = input_params
        self._routine_definition = routine_definition
        self._external_language = external_language

    @property
    def full_name(self) -> str:
        return self._full_name

    @property
    def input_params(self) -> List[FunctionParameterInfo]:
        return self._input_params

    @property
    def routine_definition(self) -> Optional[str]:
        return self._routine_definition

    @property
    def external_language(self) -> Optional[str]:
        """
        Field is None if language is SQL (not an external language).
        """
        return self._external_language

    @classmethod
    def from_dict(cls, function_info_json):
        input_params = function_info_json.get("input_params", {}).get("parameters", [])
        return FunctionInfo(
            full_name=function_info_json["full_name"],
            input_params=[FunctionParameterInfo.from_dict(p) for p in input_params],
            routine_definition=function_info_json.get("routine_definition", None),
            external_language=function_info_json.get("external_language", None),
        )


class InformationSchemaSparkClient:
    """
    Internal client to retrieve Unity Catalog metadata from system.information_schema.
    https://docs.databricks.com/sql/language-manual/sql-ref-information-schema.html
    """

    def _get_routines_with_parameters(self, full_routine_names: List[str]) -> DataFrame:
        """
        Retrieve the routines with their parameters from information_schema.routines, information_schema.parameters.
        Return DataFrame only contains routines that 1. exist and 2. the caller has GetFunction permission on.

        Note: The returned DataFrame contains the cartesian product of routines and parameters.
        For efficiency, routines table columns are only present in the first row for each routine.
        """
        routine_name_schema = StructType(
            [
                StructField("specific_catalog", StringType(), False),
                StructField("specific_schema", StringType(), False),
                StructField("specific_name", StringType(), False),
            ]
        )
        routine_names_df = self._spark_client.createDataFrame(
            [full_routine_name.split(".") for full_routine_name in full_routine_names],
            routine_name_schema,
        )
        routines_table = self._spark_client.read_table(
            "system.information_schema.routines"
        )
        parameters_table = self._spark_client.read_table(
            "system.information_schema.parameters"
        )

        # Inner join routines table to filter out non-existent routines.
        # Left join parameters as routines may have no parameters.
        full_routines_with_parameters_df = routine_names_df.join(
            routines_table, on=routine_names_df.columns, how="inner"
        ).join(parameters_table, on=routine_names_df.columns, how="left")

        # Return only relevant metadata from information_schema, sorted by routine name + parameter order.
        # For efficiency, only preserve routine column values in the first of each routine's result rows.
        # The first row will have parameter.ordinal_value is None (no parameters) or equals 0 (first parameter).
        def select_if_first_row(col: Column) -> Column:
            return when(
                isnull(parameters_table.ordinal_position)
                | (parameters_table.ordinal_position == 0),
                col,
            ).otherwise(None)

        return full_routines_with_parameters_df.select(
            routine_names_df.columns
            + [
                select_if_first_row(routines_table.routine_definition).alias(
                    "routine_definition"
                ),
                select_if_first_row(routines_table.external_language).alias(
                    "external_language"
                ),
                parameters_table.ordinal_position,
                parameters_table.parameter_name,
                parameters_table.full_data_type,
            ]
        ).sort(routine_names_df.columns + [parameters_table.ordinal_position])

    def get_functions(self, full_function_names: List[str]) -> List[FunctionInfo]:
        """
        Retrieves and maps Unity Catalog functions' metadata as FunctionInfos.
        """
        # Avoid unnecessary Spark calls and return if empty.
        if not full_function_names:
            return []

        # Collect dict of routine name -> DataFrame rows describing the routine.
        routines_with_parameters_df = self._get_routines_with_parameters(
            full_routine_names=full_function_names
        )
        routine_infos = defaultdict(list)
        for r in routines_with_parameters_df.collect():
            routine_name = f"{r.specific_catalog}.{r.specific_schema}.{r.specific_name}"
            routine_infos[routine_name].append(r)

        # Mock GetFunction DNE error, since information_schema does not throw.
        for function_name in full_function_names:
            if not function_name in routine_infos:
                raise ValueError(f"Function '{function_name}' does not exist.")

        # Map routine_infos into FunctionInfos.
        function_infos = []
        for function_name in full_function_names:
            routine_info = routine_infos[function_name][0]
            input_params = [
                FunctionParameterInfo(name=p.parameter_name, type_text=p.full_data_type)
                for p in routine_infos[function_name]
                if p.ordinal_position is not None
            ]
            function_infos.append(
                FunctionInfo(
                    full_name=function_name,
                    input_params=input_params,
                    routine_definition=routine_info.routine_definition,
                    external_language=routine_info.external_language,
                )
            )
        return function_infos



import logging

from wedata.feature_store.constants.constants import _ERROR, _WARN

_logger = logging.getLogger(__name__)


def catalog_matches_delta_schema(catalog_features, df_schema, column_filter=None):
    """
    Confirm that the column names and column types are the same.

    Returns True if identical, False if there is a mismatch.

    If column_filter is not None, only columns in column_filter must match.
    """
    if column_filter is not None:
        catalog_features = [c for c in catalog_features if c.name in column_filter]
        df_schema = [c for c in df_schema if c.name in column_filter]

    catalog_schema = {
        feature.name: feature.data_type
        for feature in catalog_features
    }
    delta_schema = {
        feature.name: feature.dataType
        for feature in df_schema
    }

    complex_catalog_schema = get_complex_catalog_schema(
        catalog_features, catalog_schema
    )
    complex_delta_schema = get_complex_delta_schema(df_schema, delta_schema)

    return (
        catalog_schema == delta_schema
        and complex_catalog_schema == complex_delta_schema
    )


def get_complex_delta_schema(delta_features, delta_feature_names_to_fs_types):
    """
    1. Filter delta features to features that have complex datatypes.
    2. Take the existing Spark DataType stored on the Delta features. This is later used for
    comparison against the Catalog schema's complex Spark DataTypes.
    3. Return a mapping of feature name to their respective complex Spark DataTypes.

    :param delta_features: List[Feature]. List of features stored in Delta.
    :param delta_feature_names_to_fs_types: Map[str, feature_store.DataType]. A mapping of feature
    names to their respective Feature Store DataTypes.
    :return: Map[str, spark.sql.types.DataType]. A mapping of feature names to their respective
    Spark DataTypes.
    """
    complex_delta_features = [
        feature
        for feature in delta_features
        if delta_feature_names_to_fs_types[feature.name] in DATA_TYPES_REQUIRES_DETAILS
    ]
    complex_delta_feature_names_to_spark_types = {
        feature.name: feature.dataType for feature in complex_delta_features
    }
    return complex_delta_feature_names_to_spark_types


def get_complex_catalog_schema(catalog_features, catalog_feature_names_to_fs_types):
    """
    1. Filter catalog features to features that have complex datatypes.
    2. Convert the JSON string stored in each feature's data_type_details to the corresponding
    Spark DataType. This is later used for comparison against the Delta schema's complex Spark
    DataTypes.
    3. Return a mapping of feature name to their respective complex Spark DataTypes.

    :param catalog_features: List[Feature]. List of features stored in the Catalog.
    :param catalog_feature_names_to_fs_types: Map[str, feature_store.DataType]. A mapping of feature
    names to their respective Feature Store DataTypes.
    :return: Map[str, spark.sql.types.DataType]. A mapping of feature names to their respective
    Spark DataTypes.
    """
    complex_catalog_features = [
        feature
        for feature in catalog_features
        if catalog_feature_names_to_fs_types[feature.name]
        in DATA_TYPES_REQUIRES_DETAILS
    ]
    complex_catalog_feature_names_to_spark_types = {
        feature.name: feature.data_type_details
        for feature in complex_catalog_features
    }
    return complex_catalog_feature_names_to_spark_types


def log_catalog_schema_not_match_delta_schema(catalog_features, df_schema, level):
    """
    Log the catalog schema does not match the delta table schema.

    Example warning:
    Expected recorded schema from Feature Catalog to be identical with
    schema in delta table.Feature Catalog's schema is
    '{'id': 'INTEGER', 'feat1': 'INTEGER'}' while delta table's
    schema is '{'id': 'INTEGER', 'feat1': 'FLOAT'}'
    """
    catalog_schema = {feature.name: feature.data_type for feature in catalog_features}
    delta_schema = {
        feature.name: feature.dataType
        for feature in df_schema
    }
    msg = (
        f"Expected recorded schema from Feature Catalog to be identical with schema "
        f"in Delta table. "
        f"Feature Catalog's schema is '{catalog_schema}' while Delta table's schema "
        f"is '{delta_schema}'"
    )
    if level == _WARN:
        _logger.warning(msg)
    elif level == _ERROR:
        raise RuntimeError(msg)
    else:
        _logger.info(msg)



import json
import re
from typing import Any

from pyspark.sql.types import ArrayType, DataType, DecimalType, MapType, StructType



class DataType(_ProtoEnumEntity):
    """Online store types."""

    INTEGER = ProtoDataType.Value("INTEGER")
    FLOAT = ProtoDataType.Value("FLOAT")
    BOOLEAN = ProtoDataType.Value("BOOLEAN")
    STRING = ProtoDataType.Value("STRING")
    DOUBLE = ProtoDataType.Value("DOUBLE")
    LONG = ProtoDataType.Value("LONG")
    TIMESTAMP = ProtoDataType.Value("TIMESTAMP")
    DATE = ProtoDataType.Value("DATE")
    SHORT = ProtoDataType.Value("SHORT")
    ARRAY = ProtoDataType.Value("ARRAY")
    MAP = ProtoDataType.Value("MAP")
    BINARY = ProtoDataType.Value("BINARY")
    DECIMAL = ProtoDataType.Value("DECIMAL")
    STRUCT = ProtoDataType.Value("STRUCT")

    _FIXED_DECIMAL = re.compile("decimal\\(\\s*(\\d+)\\s*,\\s*(\\d+)\\s*\\)")

    @classmethod
    def _enum_type(cls) -> Any:
        return ProtoDataType

    @classmethod
    def from_spark_type(cls, spark_type):
        return cls.from_string(spark_type.typeName())

    @classmethod
    def spark_type_to_string(cls, spark_type):
        return DataType.to_string(DataType.from_spark_type(spark_type))

    @classmethod
    def top_level_type_supported(cls, spark_type: DataType) -> bool:
        """
        Checks whether the provided Spark data type is supported by Feature Store, only considering
        the top-level type for nested data types.

        Details on nested types:
          ArrayType: The elementType is not checked. Will return True.
          MapType: The keyType and valueType are not checked. Will return True.
          StructType: The struct fieds are not checked. Will return True.
        """
        cls.init()
        return spark_type.typeName().upper() in cls._STRING_TO_ENUM

    @classmethod
    def to_complex_spark_type(cls, json_value):
        """
        Constructs a complex Spark DataType from its compact JSON representation.

        Examples:
            - Input: '"decimal(1,2)"'
              Output: DecimalType(1,2)
            - Input: '{"containsNull":false,"elementType":"integer","type":"array"}'
              Output: ArrayType(IntegerType,false)
            - Input: '{"keyType":"integer","type":"map","valueContainsNull":True,"valueType":"integer"}'
              Output: MapType(IntegerType,IntegerType,true)
        """
        if not json_value:
            raise ValueError("Empty JSON value cannot be converted to Spark DataType")

        json_data = json.loads(json_value)
        if not isinstance(json_data, dict):
            # DecimalType does not have fromJson() method
            if json_value == "decimal":
                return DecimalType()
            if cls._FIXED_DECIMAL.match(json_data):
                m = cls._FIXED_DECIMAL.match(json_data)
                return DecimalType(int(m.group(1)), int(m.group(2)))

        if json_data["type"].upper() == cls.to_string(cls.ARRAY):
            return ArrayType.fromJson(json_data)

        if json_data["type"].upper() == cls.to_string(cls.MAP):
            return MapType.fromJson(json_data)

        if json_data["type"].upper() == cls.to_string(cls.STRUCT):
            return StructType.fromJson(json_data)

        else:
            raise ValueError(
                f"Spark type {json_data['type']} cannot be converted to a complex Spark DataType"
            )
